// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

use std::collections::HashMap;

use super::backend::VercelArtifactsBackend;
use crate::raw::HttpClient;
use crate::Scheme;
use crate::*;

/// [Vercel Cache](https://vercel.com/docs/concepts/monorepos/remote-caching) backend support.
///
/// # Capabilities
///
/// This service can be used to:
///
/// - [ ] stat
/// - [x] read
/// - [x] write
/// - [x] create_dir
/// - [x] delete
/// - [ ] ~~copy~~
/// - [ ] ~~rename~~
/// - [ ] ~~list~~
/// - [ ] ~~scan~~
/// - [ ] ~~presign~~
/// - [ ] blocking
///
/// # Notes
/// # Configuration
///
/// - `access_token`: set the access_token for Rest API
///
/// You can refer to [`VercelArtifactsBuilder`]'s docs for more information
///
/// # Example
///
/// ## Via Builder
///
/// ```no_run
/// use anyhow::Result;
/// use opendal::services::VercelArtifacts;
/// use opendal::Operator;
///
/// #[tokio::main]
/// async fn main() -> Result<()> {
///     // create backend builder
///     let mut builder = VercelArtifacts::default();
///
///     builder.access_token("xxx");
///
///     let op: Operator = Operator::new(builder)?.finish();
///     Ok(())
/// }
/// ```
#[derive(Default)]
pub struct VercelArtifactsBuilder {
    access_token: Option<String>,
    http_client: Option<HttpClient>,
}

impl VercelArtifactsBuilder {
    /// set the bearer access token for Vercel
    ///
    /// default: no access token, which leads to failure
    pub fn access_token(&mut self, access_token: &str) -> &mut Self {
        self.access_token = Some(access_token.to_string());
        self
    }

    /// Specify the http client that used by this service.
    ///
    /// # Notes
    ///
    /// This API is part of OpenDAL's Raw API. `HttpClient` could be changed
    /// during minor updates.
    pub fn http_client(&mut self, http_client: HttpClient) -> &mut Self {
        self.http_client = Some(http_client);
        self
    }
}

impl Builder for VercelArtifactsBuilder {
    const SCHEME: Scheme = Scheme::VercelArtifacts;

    type Accessor = VercelArtifactsBackend;

    fn from_map(map: HashMap<String, String>) -> Self {
        let mut builder = Self::default();
        map.get("access_token").map(|v| builder.access_token(v));
        builder
    }

    fn build(&mut self) -> Result<Self::Accessor> {
        let client = if let Some(client) = self.http_client.take() {
            client
        } else {
            HttpClient::new().map_err(|err| {
                err.with_operation("Builder::build")
                    .with_context("service", Scheme::VercelArtifacts)
            })?
        };

        match self.access_token.clone() {
            Some(access_token) => Ok(VercelArtifactsBackend {
                access_token,
                client,
            }),
            None => Err(Error::new(ErrorKind::ConfigInvalid, "access_token not set")),
        }
    }
}
